@@ -1,76 +1,178 @@
 use crate::{
+    conversion::FromPyObjectBound,
     exceptions::PyTypeError,
     ffi,
-    type_object::PyTypeObject,
-    types::{PyDict, PyString, PyTuple},
-    FromPyObject, PyAny, PyErr, PyResult, Python,
+    pyclass::boolean_struct::False,
+    types::{any::PyAnyMethods, dict::PyDictMethods, tuple::PyTupleMethods, PyDict, PyTuple},
+    Borrowed, Bound, PyAny, PyClass, PyErr, PyRef, PyRefMut, PyResult, PyTypeCheck, Python,
 };
 
+/// Helper type used to keep implementation more concise.
+///
+/// (Function argument extraction borrows input arguments.)
+type PyArg<'py> = Borrowed<'py, 'py, PyAny>;
+
+/// A trait which is used to help PyO3 macros extract function arguments.
+///
+/// `#[pyclass]` structs need to extract as `PyRef<T>` and `PyRefMut<T>`
+/// wrappers rather than extracting `&T` and `&mut T` directly. The `Holder` type is used
+/// to hold these temporary wrappers - the way the macro is constructed, these wrappers
+/// will be dropped as soon as the pyfunction call ends.
+///
+/// There exists a trivial blanket implementation for `T: FromPyObject` with `Holder = ()`.
+pub trait PyFunctionArgument<'a, 'py, const IS_OPTION: bool>: Sized + 'a {
+    type Holder: FunctionArgumentHolder;
+    fn extract(obj: &'a Bound<'py, PyAny>, holder: &'a mut Self::Holder) -> PyResult<Self>;
+}
+
+impl<'a, 'py, T> PyFunctionArgument<'a, 'py, false> for T
+where
+    T: FromPyObjectBound<'a, 'py> + 'a,
+{
+    type Holder = ();
+
+    #[inline]
+    fn extract(obj: &'a Bound<'py, PyAny>, _: &'a mut ()) -> PyResult<Self> {
+        obj.extract()
+    }
+}
+
+impl<'a, 'py, T: 'py> PyFunctionArgument<'a, 'py, false> for &'a Bound<'py, T>
+where
+    T: PyTypeCheck,
+{
+    type Holder = ();
+
+    #[inline]
+    fn extract(obj: &'a Bound<'py, PyAny>, _: &'a mut ()) -> PyResult<Self> {
+        obj.downcast().map_err(Into::into)
+    }
+}
+
+impl<'a, 'py, T> PyFunctionArgument<'a, 'py, true> for Option<T>
+where
+    T: PyFunctionArgument<'a, 'py, false>, // inner `Option`s will use `FromPyObject`
+{
+    type Holder = T::Holder;
+
+    #[inline]
+    fn extract(obj: &'a Bound<'py, PyAny>, holder: &'a mut T::Holder) -> PyResult<Self> {
+        if obj.is_none() {
+            Ok(None)
+        } else {
+            Ok(Some(T::extract(obj, holder)?))
+        }
+    }
+}
+
+#[cfg(all(Py_LIMITED_API, not(Py_3_10)))]
+impl<'a> PyFunctionArgument<'a, '_, false> for &'a str {
+    type Holder = Option<std::borrow::Cow<'a, str>>;
+
+    #[inline]
+    fn extract(
+        obj: &'a Bound<'_, PyAny>,
+        holder: &'a mut Option<std::borrow::Cow<'a, str>>,
+    ) -> PyResult<Self> {
+        Ok(holder.insert(obj.extract()?))
+    }
+}
+
+/// Trait for types which can be a function argument holder - they should
+/// to be able to const-initialize to an empty value.
+pub trait FunctionArgumentHolder: Sized {
+    const INIT: Self;
+}
+
+impl FunctionArgumentHolder for () {
+    const INIT: Self = ();
+}
+
+impl<T> FunctionArgumentHolder for Option<T> {
+    const INIT: Self = None;
+}
+
+#[inline]
+pub fn extract_pyclass_ref<'a, 'py: 'a, T: PyClass>(
+    obj: &'a Bound<'py, PyAny>,
+    holder: &'a mut Option<PyRef<'py, T>>,
+) -> PyResult<&'a T> {
+    Ok(&*holder.insert(obj.extract()?))
+}
+
+#[inline]
+pub fn extract_pyclass_ref_mut<'a, 'py: 'a, T: PyClass<Frozen = False>>(
+    obj: &'a Bound<'py, PyAny>,
+    holder: &'a mut Option<PyRefMut<'py, T>>,
+) -> PyResult<&'a mut T> {
+    Ok(&mut *holder.insert(obj.extract()?))
+}
+
 /// The standard implementation of how PyO3 extracts a `#[pyfunction]` or `#[pymethod]` function argument.
 #[doc(hidden)]
-#[inline]
-pub fn extract_argument<'py, T>(obj: &'py PyAny, arg_name: &str) -> PyResult<T>
+pub fn extract_argument<'a, 'py, T, const IS_OPTION: bool>(
+    obj: &'a Bound<'py, PyAny>,
+    holder: &'a mut T::Holder,
+    arg_name: &str,
+) -> PyResult<T>
 where
-    T: FromPyObject<'py>,
+    T: PyFunctionArgument<'a, 'py, IS_OPTION>,
 {
-    match obj.extract() {
+    match PyFunctionArgument::extract(obj, holder) {
         Ok(value) => Ok(value),
         Err(e) => Err(argument_extraction_error(obj.py(), arg_name, e)),
     }
 }
 
-/// Alternative to [`extract_argument`] used for `Option<T>` arguments (because they are implicitly treated
-/// as optional if at the end of the positional parameters).
+/// Alternative to [`extract_argument`] used for `Option<T>` arguments. This is necessary because Option<&T>
+/// does not implement `PyFunctionArgument` for `T: PyClass`.
 #[doc(hidden)]
-#[inline]
-pub fn extract_optional_argument<'py, T>(
-    obj: Option<&'py PyAny>,
+pub fn extract_optional_argument<'a, 'py, T, const IS_OPTION: bool>(
+    obj: Option<&'a Bound<'py, PyAny>>,
+    holder: &'a mut T::Holder,
     arg_name: &str,
+    default: fn() -> Option<T>,
 ) -> PyResult<Option<T>>
 where
-    T: FromPyObject<'py>,
+    T: PyFunctionArgument<'a, 'py, IS_OPTION>,
 {
     match obj {
-        Some(obj) => match obj.extract() {
-            Ok(value) => Ok(value),
-            Err(e) => Err(argument_extraction_error(obj.py(), arg_name, e)),
-        },
-        None => Ok(None),
+        Some(obj) => {
+            if obj.is_none() {
+                // Explicit `None` will result in None being used as the function argument
+                Ok(None)
+            } else {
+                extract_argument(obj, holder, arg_name).map(Some)
+            }
+        }
+        _ => Ok(default()),
     }
 }
 
 /// Alternative to [`extract_argument`] used when the argument has a default value provided by an annotation.
 #[doc(hidden)]
-#[inline]
-pub fn extract_argument_with_default<'py, T>(
-    obj: Option<&'py PyAny>,
+pub fn extract_argument_with_default<'a, 'py, T, const IS_OPTION: bool>(
+    obj: Option<&'a Bound<'py, PyAny>>,
+    holder: &'a mut T::Holder,
     arg_name: &str,
-    default: impl FnOnce() -> T,
+    default: fn() -> T,
 ) -> PyResult<T>
 where
-    T: FromPyObject<'py>,
+    T: PyFunctionArgument<'a, 'py, IS_OPTION>,
 {
     match obj {
-        Some(obj) => match obj.extract() {
-            Ok(value) => Ok(value),
-            Err(e) => Err(argument_extraction_error(obj.py(), arg_name, e)),
-        },
+        Some(obj) => extract_argument(obj, holder, arg_name),
         None => Ok(default()),
     }
 }
 
 /// Alternative to [`extract_argument`] used when the argument has a `#[pyo3(from_py_with)]` annotation.
-///
-/// # Safety
-/// - `obj` must not be None (this helper is only used for required function arguments).
 #[doc(hidden)]
-#[inline]
-pub fn from_py_with<'py, T>(
-    obj: &'py PyAny,
+pub fn from_py_with<'a, 'py, T>(
+    obj: &'a Bound<'py, PyAny>,
     arg_name: &str,
-    extractor: impl FnOnce(&'py PyAny) -> PyResult<T>,
+    extractor: fn(&'a Bound<'py, PyAny>) -> PyResult<T>,
 ) -> PyResult<T> {
-    // Safety: obj is not None (see safety
     match extractor(obj) {
         Ok(value) => Ok(value),
         Err(e) => Err(argument_extraction_error(obj.py(), arg_name, e)),
@@ -79,18 +181,14 @@ pub fn from_py_with<'py, T>(
 
 /// Alternative to [`extract_argument`] used when the argument has a `#[pyo3(from_py_with)]` annotation and also a default value.
 #[doc(hidden)]
-#[inline]
-pub fn from_py_with_with_default<'py, T>(
-    obj: Option<&'py PyAny>,
+pub fn from_py_with_with_default<'a, 'py, T>(
+    obj: Option<&'a Bound<'py, PyAny>>,
     arg_name: &str,
-    extractor: impl FnOnce(&'py PyAny) -> PyResult<T>,
-    default: impl FnOnce() -> T,
+    extractor: fn(&'a Bound<'py, PyAny>) -> PyResult<T>,
+    default: fn() -> T,
 ) -> PyResult<T> {
     match obj {
-        Some(obj) => match extractor(obj) {
-            Ok(value) => Ok(value),
-            Err(e) => Err(argument_extraction_error(obj.py(), arg_name, e)),
-        },
+        Some(obj) => from_py_with(obj, arg_name, extractor),
         None => Ok(default()),
     }
 }
@@ -101,8 +199,8 @@ pub fn from_py_with_with_default<'py, T>(
 /// single string.)
 #[doc(hidden)]
 #[cold]
-pub fn argument_extraction_error(py: Python, arg_name: &str, error: PyErr) -> PyErr {
-    if error.get_type(py).is(PyTypeError::type_object(py)) {
+pub fn argument_extraction_error(py: Python<'_>, arg_name: &str, error: PyErr) -> PyErr {
+    if error.get_type(py).is(&py.get_type::<PyTypeError>()) {
         let remapped_error =
             PyTypeError::new_err(format!("argument '{}': {}", arg_name, error.value(py)));
         remapped_error.set_cause(py, error.cause(py));
@@ -119,7 +217,9 @@ pub fn argument_extraction_error(py: Python, arg_name: &str, error: PyErr) -> Py
 /// `argument` must not be `None`
 #[doc(hidden)]
 #[inline]
-pub unsafe fn unwrap_required_argument(argument: Option<&PyAny>) -> &PyAny {
+pub unsafe fn unwrap_required_argument<'a, 'py>(
+    argument: Option<&'a Bound<'py, PyAny>>,
+) -> &'a Bound<'py, PyAny> {
     match argument {
         Some(value) => value,
         #[cfg(debug_assertions)]
@@ -156,29 +256,25 @@ impl FunctionDescription {
     /// Equivalent of `extract_arguments_tuple_dict` which uses the Python C-API "fastcall" convention.
     ///
     /// # Safety
-    /// - `args` must be a pointer to a C-style array of valid `ffi::PyObject` pointers.
+    /// - `args` must be a pointer to a C-style array of valid `ffi::PyObject` pointers, or NULL.
     /// - `kwnames` must be a pointer to a PyTuple, or NULL.
     /// - `nargs + kwnames.len()` is the total length of the `args` array.
-    #[cfg(not(Py_LIMITED_API))]
+    #[cfg(any(Py_3_10, not(Py_LIMITED_API)))]
     pub unsafe fn extract_arguments_fastcall<'py, V, K>(
         &self,
         py: Python<'py>,
         args: *const *mut ffi::PyObject,
         nargs: ffi::Py_ssize_t,
         kwnames: *mut ffi::PyObject,
-        output: &mut [Option<&'py PyAny>],
+        output: &mut [Option<PyArg<'py>>],
     ) -> PyResult<(V::Varargs, K::Varkeywords)>
     where
         V: VarargsHandler<'py>,
         K: VarkeywordsHandler<'py>,
     {
-        // Safety: Option<&PyAny> has the same memory layout as `*mut ffi::PyObject`
-        let args = args as *const Option<&PyAny>;
-        let positional_args_provided = nargs as usize;
-        let args_slice = std::slice::from_raw_parts(args, positional_args_provided);
-
         let num_positional_parameters = self.positional_parameter_names.len();
 
+        debug_assert!(nargs >= 0);
         debug_assert!(self.positional_only_parameters <= num_positional_parameters);
         debug_assert!(self.required_positional_parameters <= num_positional_parameters);
         debug_assert_eq!(
@@ -186,57 +282,52 @@ impl FunctionDescription {
             num_positional_parameters + self.keyword_only_parameters.len()
         );
 
-        let varargs = if positional_args_provided > num_positional_parameters {
-            let (positional_parameters, varargs) = args_slice.split_at(num_positional_parameters);
-            output[..num_positional_parameters].copy_from_slice(positional_parameters);
-            V::handle_varargs_fastcall(py, varargs, self)?
+        // Handle positional arguments
+        // Safety:
+        //  - Option<PyArg> has the same memory layout as `*mut ffi::PyObject`
+        //  - we both have the GIL and can borrow these input references for the `'py` lifetime.
+        let args: *const Option<PyArg<'py>> = args.cast();
+        let positional_args_provided = nargs as usize;
+        let remaining_positional_args = if args.is_null() {
+            debug_assert_eq!(positional_args_provided, 0);
+            &[]
         } else {
-            output[..positional_args_provided].copy_from_slice(args_slice);
-            V::handle_varargs_fastcall(py, &[], self)?
+            // Can consume at most the number of positional parameters in the function definition,
+            // the rest are varargs.
+            let positional_args_to_consume =
+                num_positional_parameters.min(positional_args_provided);
+            let (positional_parameters, remaining) = unsafe {
+                std::slice::from_raw_parts(args, positional_args_provided)
+                    .split_at(positional_args_to_consume)
+            };
+            output[..positional_args_to_consume].copy_from_slice(positional_parameters);
+            remaining
         };
+        let varargs = V::handle_varargs_fastcall(py, remaining_positional_args, self)?;
 
         // Handle keyword arguments
-        let mut varkeywords = Default::default();
-        if let Some(kwnames) = py.from_borrowed_ptr_or_opt::<PyTuple>(kwnames) {
-            let mut positional_only_keyword_arguments = Vec::new();
+        let mut varkeywords = K::Varkeywords::default();
 
-            // Safety: &PyAny has the same memory layout as `*mut ffi::PyObject`
-            let kwargs =
-                ::std::slice::from_raw_parts((args as *const &PyAny).offset(nargs), kwnames.len());
+        // Safety: kwnames is known to be a pointer to a tuple, or null
+        //  - we both have the GIL and can borrow this input reference for the `'py` lifetime.
+        let kwnames: Option<Borrowed<'_, '_, PyTuple>> = unsafe {
+            Borrowed::from_ptr_or_opt(py, kwnames).map(|kwnames| kwnames.downcast_unchecked())
+        };
+        if let Some(kwnames) = kwnames {
+            let kwargs = unsafe {
+                ::std::slice::from_raw_parts(
+                    // Safety: PyArg has the same memory layout as `*mut ffi::PyObject`
+                    args.offset(nargs).cast::<PyArg<'py>>(),
+                    kwnames.len(),
+                )
+            };
 
-            for (kwarg_name_py, &value) in kwnames.iter().zip(kwargs) {
-                // All keyword arguments should be UTF8 strings, but we'll check, just in case.
-                if let Ok(kwarg_name) = kwarg_name_py.downcast::<PyString>()?.to_str() {
-                    // Try to place parameter in keyword only parameters
-                    if let Some(i) = self.find_keyword_parameter_in_keyword_only(kwarg_name) {
-                        if output[i + num_positional_parameters]
-                            .replace(value)
-                            .is_some()
-                        {
-                            return Err(self.multiple_values_for_argument(kwarg_name));
-                        }
-                        continue;
-                    }
-
-                    // Repeat for positional parameters
-                    if let Some(i) = self.find_keyword_parameter_in_positional(kwarg_name) {
-                        if i < self.positional_only_parameters {
-                            positional_only_keyword_arguments.push(kwarg_name);
-                        } else if output[i].replace(value).is_some() {
-                            return Err(self.multiple_values_for_argument(kwarg_name));
-                        }
-                        continue;
-                    }
-                };
-
-                K::handle_unexpected_keyword(&mut varkeywords, kwarg_name_py, value, self)?
-            }
-
-            if !positional_only_keyword_arguments.is_empty() {
-                return Err(
-                    self.positional_only_keyword_arguments(&positional_only_keyword_arguments)
-                );
-            }
+            self.handle_kwargs::<K, _>(
+                kwnames.iter_borrowed().zip(kwargs.iter().copied()),
+                &mut varkeywords,
+                num_positional_parameters,
+                output,
+            )?
         }
 
         // Once all inputs have been processed, check that all required arguments have been provided.
@@ -264,14 +355,21 @@ impl FunctionDescription {
         py: Python<'py>,
         args: *mut ffi::PyObject,
         kwargs: *mut ffi::PyObject,
-        output: &mut [Option<&'py PyAny>],
+        output: &mut [Option<PyArg<'py>>],
     ) -> PyResult<(V::Varargs, K::Varkeywords)>
     where
         V: VarargsHandler<'py>,
         K: VarkeywordsHandler<'py>,
     {
-        let args = py.from_borrowed_ptr::<PyTuple>(args);
-        let kwargs: ::std::option::Option<&PyDict> = py.from_borrowed_ptr_or_opt(kwargs);
+        // Safety:
+        //  - `args` is known to be a tuple
+        //  - `kwargs` is known to be a dict or null
+        //  - we both have the GIL and can borrow these input references for the `'py` lifetime.
+        let args: Borrowed<'py, 'py, PyTuple> =
+            unsafe { Borrowed::from_ptr(py, args).downcast_unchecked::<PyTuple>() };
+        let kwargs: Option<Borrowed<'py, 'py, PyDict>> = unsafe {
+            Borrowed::from_ptr_or_opt(py, kwargs).map(|kwargs| kwargs.downcast_unchecked())
+        };
 
         let num_positional_parameters = self.positional_parameter_names.len();
 
@@ -283,50 +381,26 @@ impl FunctionDescription {
         );
 
         // Copy positional arguments into output
-        for (i, arg) in args.iter().take(num_positional_parameters).enumerate() {
+        for (i, arg) in args
+            .iter_borrowed()
+            .take(num_positional_parameters)
+            .enumerate()
+        {
             output[i] = Some(arg);
         }
 
         // If any arguments remain, push them to varargs (if possible) or error
-        let varargs = V::handle_varargs_tuple(args, self)?;
+        let varargs = V::handle_varargs_tuple(&args, self)?;
 
         // Handle keyword arguments
-        let mut varkeywords = Default::default();
+        let mut varkeywords = K::Varkeywords::default();
         if let Some(kwargs) = kwargs {
-            let mut positional_only_keyword_arguments = Vec::new();
-            for (kwarg_name_py, value) in kwargs {
-                // All keyword arguments should be UTF8 strings, but we'll check, just in case.
-                if let Ok(kwarg_name) = kwarg_name_py.downcast::<PyString>()?.to_str() {
-                    // Try to place parameter in keyword only parameters
-                    if let Some(i) = self.find_keyword_parameter_in_keyword_only(kwarg_name) {
-                        if output[i + num_positional_parameters]
-                            .replace(value)
-                            .is_some()
-                        {
-                            return Err(self.multiple_values_for_argument(kwarg_name));
-                        }
-                        continue;
-                    }
-
-                    // Repeat for positional parameters
-                    if let Some(i) = self.find_keyword_parameter_in_positional(kwarg_name) {
-                        if i < self.positional_only_parameters {
-                            positional_only_keyword_arguments.push(kwarg_name);
-                        } else if output[i].replace(value).is_some() {
-                            return Err(self.multiple_values_for_argument(kwarg_name));
-                        }
-                        continue;
-                    }
-                };
-
-                K::handle_unexpected_keyword(&mut varkeywords, kwarg_name_py, value, self)?
-            }
-
-            if !positional_only_keyword_arguments.is_empty() {
-                return Err(
-                    self.positional_only_keyword_arguments(&positional_only_keyword_arguments)
-                );
-            }
+            self.handle_kwargs::<K, _>(
+                unsafe { kwargs.iter_borrowed() },
+                &mut varkeywords,
+                num_positional_parameters,
+                output,
+            )?
         }
 
         // Once all inputs have been processed, check that all required arguments have been provided.
@@ -337,6 +411,85 @@ impl FunctionDescription {
         Ok((varargs, varkeywords))
     }
 
+    #[inline]
+    fn handle_kwargs<'py, K, I>(
+        &self,
+        kwargs: I,
+        varkeywords: &mut K::Varkeywords,
+        num_positional_parameters: usize,
+        output: &mut [Option<PyArg<'py>>],
+    ) -> PyResult<()>
+    where
+        K: VarkeywordsHandler<'py>,
+        I: IntoIterator<Item = (PyArg<'py>, PyArg<'py>)>,
+    {
+        debug_assert_eq!(
+            num_positional_parameters,
+            self.positional_parameter_names.len()
+        );
+        debug_assert_eq!(
+            output.len(),
+            num_positional_parameters + self.keyword_only_parameters.len()
+        );
+        let mut positional_only_keyword_arguments = Vec::new();
+        for (kwarg_name_py, value) in kwargs {
+            // Safety: All keyword arguments should be UTF-8 strings, but if it's not, `.to_str()`
+            // will return an error anyway.
+            #[cfg(any(Py_3_10, not(Py_LIMITED_API)))]
+            let kwarg_name =
+                unsafe { kwarg_name_py.downcast_unchecked::<crate::types::PyString>() }.to_str();
+
+            #[cfg(all(not(Py_3_10), Py_LIMITED_API))]
+            let kwarg_name = kwarg_name_py.extract::<crate::pybacked::PyBackedStr>();
+
+            if let Ok(kwarg_name_owned) = kwarg_name {
+                #[cfg(any(Py_3_10, not(Py_LIMITED_API)))]
+                let kwarg_name = kwarg_name_owned;
+                #[cfg(all(not(Py_3_10), Py_LIMITED_API))]
+                let kwarg_name: &str = &kwarg_name_owned;
+
+                // Try to place parameter in keyword only parameters
+                if let Some(i) = self.find_keyword_parameter_in_keyword_only(kwarg_name) {
+                    if output[i + num_positional_parameters]
+                        .replace(value)
+                        .is_some()
+                    {
+                        return Err(self.multiple_values_for_argument(kwarg_name));
+                    }
+                    continue;
+                }
+
+                // Repeat for positional parameters
+                if let Some(i) = self.find_keyword_parameter_in_positional(kwarg_name) {
+                    if i < self.positional_only_parameters {
+                        // If accepting **kwargs, then it's allowed for the name of the
+                        // kwarg to conflict with a postional-only argument - the value
+                        // will go into **kwargs anyway.
+                        if K::handle_varkeyword(varkeywords, kwarg_name_py, value, self).is_err() {
+                            positional_only_keyword_arguments.push(kwarg_name_owned);
+                        }
+                    } else if output[i].replace(value).is_some() {
+                        return Err(self.multiple_values_for_argument(kwarg_name));
+                    }
+                    continue;
+                }
+            };
+
+            K::handle_varkeyword(varkeywords, kwarg_name_py, value, self)?
+        }
+
+        if !positional_only_keyword_arguments.is_empty() {
+            #[cfg(all(not(Py_3_10), Py_LIMITED_API))]
+            let positional_only_keyword_arguments: Vec<_> = positional_only_keyword_arguments
+                .iter()
+                .map(std::ops::Deref::deref)
+                .collect();
+            return Err(self.positional_only_keyword_arguments(&positional_only_keyword_arguments));
+        }
+
+        Ok(())
+    }
+
     #[inline]
     fn find_keyword_parameter_in_positional(&self, kwarg_name: &str) -> Option<usize> {
         self.positional_parameter_names
@@ -357,7 +510,7 @@ impl FunctionDescription {
     #[inline]
     fn ensure_no_missing_required_positional_arguments(
         &self,
-        output: &[Option<&PyAny>],
+        output: &[Option<PyArg<'_>>],
         positional_args_provided: usize,
     ) -> PyResult<()> {
         if positional_args_provided < self.required_positional_parameters {
@@ -373,7 +526,7 @@ impl FunctionDescription {
     #[inline]
     fn ensure_no_missing_required_keyword_arguments(
         &self,
-        output: &[Option<&PyAny>],
+        output: &[Option<PyArg<'_>>],
     ) -> PyResult<()> {
         let keyword_output = &output[self.positional_parameter_names.len()..];
         for (param, out) in self.keyword_only_parameters.iter().zip(keyword_output) {
@@ -418,11 +571,11 @@ impl FunctionDescription {
     }
 
     #[cold]
-    fn unexpected_keyword_argument(&self, argument: &PyAny) -> PyErr {
+    fn unexpected_keyword_argument(&self, argument: PyArg<'_>) -> PyErr {
         PyTypeError::new_err(format!(
             "{} got an unexpected keyword argument '{}'",
             self.full_name(),
-            argument
+            argument.as_any()
         ))
     }
 
@@ -455,7 +608,7 @@ impl FunctionDescription {
     }
 
     #[cold]
-    fn missing_required_keyword_arguments(&self, keyword_outputs: &[Option<&PyAny>]) -> PyErr {
+    fn missing_required_keyword_arguments(&self, keyword_outputs: &[Option<PyArg<'_>>]) -> PyErr {
         debug_assert_eq!(self.keyword_only_parameters.len(), keyword_outputs.len());
 
         let missing_keyword_only_arguments: Vec<_> = self
@@ -476,7 +629,7 @@ impl FunctionDescription {
     }
 
     #[cold]
-    fn missing_required_positional_arguments(&self, output: &[Option<&PyAny>]) -> PyErr {
+    fn missing_required_positional_arguments(&self, output: &[Option<PyArg<'_>>]) -> PyErr {
         let missing_positional_arguments: Vec<_> = self
             .positional_parameter_names
             .iter()
@@ -496,14 +649,14 @@ pub trait VarargsHandler<'py> {
     /// Called by `FunctionDescription::extract_arguments_fastcall` with any additional arguments.
     fn handle_varargs_fastcall(
         py: Python<'py>,
-        varargs: &[Option<&PyAny>],
+        varargs: &[Option<PyArg<'py>>],
         function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs>;
     /// Called by `FunctionDescription::extract_arguments_tuple_dict` with the original tuple.
     ///
     /// Additional arguments are those in the tuple slice starting from `function_description.positional_parameter_names.len()`.
     fn handle_varargs_tuple(
-        args: &'py PyTuple,
+        args: &Bound<'py, PyTuple>,
         function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs>;
 }
@@ -517,7 +670,7 @@ impl<'py> VarargsHandler<'py> for NoVarargs {
     #[inline]
     fn handle_varargs_fastcall(
         _py: Python<'py>,
-        varargs: &[Option<&PyAny>],
+        varargs: &[Option<PyArg<'py>>],
         function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs> {
         let extra_arguments = varargs.len();
@@ -531,7 +684,7 @@ impl<'py> VarargsHandler<'py> for NoVarargs {
 
     #[inline]
     fn handle_varargs_tuple(
-        args: &'py PyTuple,
+        args: &Bound<'py, PyTuple>,
         function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs> {
         let positional_parameter_count = function_description.positional_parameter_names.len();
@@ -548,19 +701,19 @@ impl<'py> VarargsHandler<'py> for NoVarargs {
 pub struct TupleVarargs;
 
 impl<'py> VarargsHandler<'py> for TupleVarargs {
-    type Varargs = &'py PyTuple;
+    type Varargs = Bound<'py, PyTuple>;
     #[inline]
     fn handle_varargs_fastcall(
         py: Python<'py>,
-        varargs: &[Option<&PyAny>],
+        varargs: &[Option<PyArg<'py>>],
         _function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs> {
-        Ok(PyTuple::new(py, varargs))
+        PyTuple::new(py, varargs)
     }
 
     #[inline]
     fn handle_varargs_tuple(
-        args: &'py PyTuple,
+        args: &Bound<'py, PyTuple>,
         function_description: &FunctionDescription,
     ) -> PyResult<Self::Varargs> {
         let positional_parameters = function_description.positional_parameter_names.len();
@@ -568,13 +721,13 @@ impl<'py> VarargsHandler<'py> for TupleVarargs {
     }
 }
 
-/// A trait used to control whether to accept unrecognised keywords in FunctionDescription::extract_argument_(method) functions.
+/// A trait used to control whether to accept varkeywords in FunctionDescription::extract_argument_(method) functions.
 pub trait VarkeywordsHandler<'py> {
     type Varkeywords: Default;
-    fn handle_unexpected_keyword(
+    fn handle_varkeyword(
         varkeywords: &mut Self::Varkeywords,
-        name: &'py PyAny,
-        value: &'py PyAny,
+        name: PyArg<'py>,
+        value: PyArg<'py>,
         function_description: &FunctionDescription,
     ) -> PyResult<()>;
 }
@@ -585,10 +738,10 @@ pub struct NoVarkeywords;
 impl<'py> VarkeywordsHandler<'py> for NoVarkeywords {
     type Varkeywords = ();
     #[inline]
-    fn handle_unexpected_keyword(
+    fn handle_varkeyword(
         _varkeywords: &mut Self::Varkeywords,
-        name: &'py PyAny,
-        _value: &'py PyAny,
+        name: PyArg<'py>,
+        _value: PyArg<'py>,
         function_description: &FunctionDescription,
     ) -> PyResult<()> {
         Err(function_description.unexpected_keyword_argument(name))
@@ -599,12 +752,12 @@ impl<'py> VarkeywordsHandler<'py> for NoVarkeywords {
 pub struct DictVarkeywords;
 
 impl<'py> VarkeywordsHandler<'py> for DictVarkeywords {
-    type Varkeywords = Option<&'py PyDict>;
+    type Varkeywords = Option<Bound<'py, PyDict>>;
     #[inline]
-    fn handle_unexpected_keyword(
+    fn handle_varkeyword(
         varkeywords: &mut Self::Varkeywords,
-        name: &'py PyAny,
-        value: &'py PyAny,
+        name: PyArg<'py>,
+        value: PyArg<'py>,
         _function_description: &FunctionDescription,
     ) -> PyResult<()> {
         varkeywords
@@ -614,13 +767,14 @@ impl<'py> VarkeywordsHandler<'py> for DictVarkeywords {
 }
 
 fn push_parameter_list(msg: &mut String, parameter_names: &[&str]) {
+    let len = parameter_names.len();
     for (i, parameter) in parameter_names.iter().enumerate() {
         if i != 0 {
-            if parameter_names.len() > 2 {
+            if len > 2 {
                 msg.push(',');
             }
 
-            if i == parameter_names.len() - 1 {
+            if i == len - 1 {
                 msg.push_str(" and ")
             } else {
                 msg.push(' ')
@@ -635,10 +789,8 @@ fn push_parameter_list(msg: &mut String, parameter_names: &[&str]) {
 
 #[cfg(test)]
 mod tests {
-    use crate::{
-        types::{IntoPyDict, PyTuple},
-        AsPyPointer, PyAny, Python, ToPyObject,
-    };
+    use crate::types::{IntoPyDict, PyTuple};
+    use crate::Python;
 
     use super::{push_parameter_list, FunctionDescription, NoVarargs, NoVarkeywords};
 
@@ -654,14 +806,14 @@ mod tests {
         };
 
         Python::with_gil(|py| {
+            let args = PyTuple::empty(py);
+            let kwargs = [("foo", 0u8)].into_py_dict(py).unwrap();
             let err = unsafe {
                 function_description
                     .extract_arguments_tuple_dict::<NoVarargs, NoVarkeywords>(
                         py,
-                        PyTuple::new(py, Vec::<&PyAny>::new()).as_ptr(),
-                        [("foo".to_object(py).into_ref(py), 0u8)]
-                            .into_py_dict(py)
-                            .as_ptr(),
+                        args.as_ptr(),
+                        kwargs.as_ptr(),
                         &mut [],
                     )
                     .unwrap_err()
@@ -685,21 +837,21 @@ mod tests {
         };
 
         Python::with_gil(|py| {
+            let args = PyTuple::empty(py);
+            let kwargs = [(1u8, 1u8)].into_py_dict(py).unwrap();
             let err = unsafe {
                 function_description
                     .extract_arguments_tuple_dict::<NoVarargs, NoVarkeywords>(
                         py,
-                        PyTuple::new(py, Vec::<&PyAny>::new()).as_ptr(),
-                        [(1u8.to_object(py).into_ref(py), 1u8)]
-                            .into_py_dict(py)
-                            .as_ptr(),
+                        args.as_ptr(),
+                        kwargs.as_ptr(),
                         &mut [],
                     )
                     .unwrap_err()
             };
             assert_eq!(
                 err.to_string(),
-                "TypeError: 'int' object cannot be converted to 'PyString'"
+                "TypeError: example() got an unexpected keyword argument '1'"
             );
         })
     }
@@ -716,11 +868,12 @@ mod tests {
         };
 
         Python::with_gil(|py| {
+            let args = PyTuple::empty(py);
             let mut output = [None, None];
             let err = unsafe {
                 function_description.extract_arguments_tuple_dict::<NoVarargs, NoVarkeywords>(
                     py,
-                    PyTuple::new(py, Vec::<&PyAny>::new()).as_ptr(),
+                    args.as_ptr(),
                     std::ptr::null_mut(),
                     &mut output,
                 )
