@@ -422,6 +422,7 @@ mod tests {
         use super::*;
         use crate::ffi;
         use crate::{py_result_ext::PyResultExt, types::PyType};
+        use std::ptr;
 
         fn get_type(py: Python<'_>) -> PyResult<Bound<'_, PyType>> {
             py.run(ffi::c_str!("class A:\n    pass\n"), None, None)?;
@@ -450,8 +451,10 @@ mod tests {
                         let obj = obj.unwrap();
 
                         assert!(obj.is_some());
-                        assert!(obj.map_or(false, |obj| obj.as_ptr() == object.as_ptr()
-                            && obj.is_exact_instance(&class)));
+                        assert!(
+                            obj.map_or(false, |obj| ptr::eq(obj.as_ptr(), object.as_ptr())
+                                && obj.is_exact_instance(&class))
+                        );
                     }
 
                     drop(object);
@@ -492,8 +495,10 @@ mod tests {
                         let obj = unsafe { reference.upgrade_as_unchecked::<PyAny>() };
 
                         assert!(obj.is_some());
-                        assert!(obj.map_or(false, |obj| obj.as_ptr() == object.as_ptr()
-                            && obj.is_exact_instance(&class)));
+                        assert!(
+                            obj.map_or(false, |obj| ptr::eq(obj.as_ptr(), object.as_ptr())
+                                && obj.is_exact_instance(&class))
+                        );
                     }
 
                     drop(object);
@@ -586,6 +591,7 @@ mod tests {
     mod pyo3_pyclass {
         use super::*;
         use crate::{pyclass, Py};
+        use std::ptr;
 
         #[pyclass(weakref, crate = "crate")]
         struct WeakrefablePyClass {}
@@ -609,7 +615,7 @@ mod tests {
                         let obj = obj.unwrap();
 
                         assert!(obj.is_some());
-                        assert!(obj.map_or(false, |obj| obj.as_ptr() == object.as_ptr()));
+                        assert!(obj.map_or(false, |obj| ptr::eq(obj.as_ptr(), object.as_ptr())));
                     }
 
                     drop(object);
@@ -647,7 +653,7 @@ mod tests {
                         let obj = unsafe { reference.upgrade_as_unchecked::<WeakrefablePyClass>() };
 
                         assert!(obj.is_some());
-                        assert!(obj.map_or(false, |obj| obj.as_ptr() == object.as_ptr()));
+                        assert!(obj.map_or(false, |obj| ptr::eq(obj.as_ptr(), object.as_ptr())));
                     }
 
                     drop(object);
