@@ -1,4 +1,5 @@
 #![cfg(feature = "macros")]
+#![cfg_attr(not(cargo_toml_lints), warn(unsafe_op_in_unsafe_fn))]
 
 use std::collections::HashMap;
 
@@ -45,6 +46,45 @@ fn test_optional_bool() {
     });
 }
 
+#[pyfunction]
+#[pyo3(signature=(arg))]
+fn required_optional_str(arg: Option<&str>) -> &str {
+    arg.unwrap_or("")
+}
+
+#[test]
+fn test_optional_str() {
+    // Regression test for issue #4965
+    Python::with_gil(|py| {
+        let f = wrap_pyfunction!(required_optional_str)(py).unwrap();
+
+        py_assert!(py, f, "f('') == ''");
+        py_assert!(py, f, "f('foo') == 'foo'");
+        py_assert!(py, f, "f(None) == ''");
+    });
+}
+
+#[pyclass]
+struct MyClass();
+
+#[pyfunction]
+#[pyo3(signature=(arg))]
+fn required_optional_class(arg: Option<&MyClass>) {
+    let _ = arg;
+}
+
+#[test]
+fn test_required_optional_class() {
+    // Regression test for issue #4965
+    Python::with_gil(|py| {
+        let f = wrap_pyfunction!(required_optional_class)(py).unwrap();
+        let val = Bound::new(py, MyClass()).unwrap();
+
+        py_assert!(py, f val, "f(val) is None");
+        py_assert!(py, f, "f(None) is None");
+    });
+}
+
 #[cfg(not(Py_LIMITED_API))]
 #[pyfunction]
 fn buffer_inplace_add(py: Python<'_>, x: PyBuffer<i32>, y: PyBuffer<i32>) {
@@ -339,7 +379,7 @@ fn test_pycfunction_new() {
             _self: *mut ffi::PyObject,
             _args: *mut ffi::PyObject,
         ) -> *mut ffi::PyObject {
-            ffi::PyLong_FromLong(4200)
+            unsafe { ffi::PyLong_FromLong(4200) }
         }
 
         let py_fn = PyCFunction::new(
@@ -389,24 +429,26 @@ fn test_pycfunction_new_with_keywords() {
                 ptr::null_mut(),
             ];
 
-            ffi::PyArg_ParseTupleAndKeywords(
-                args,
-                kwds,
-                c_str!("l|l").as_ptr(),
-                #[cfg(Py_3_13)]
-                args_names.as_ptr(),
-                #[cfg(not(Py_3_13))]
-                args_names.as_mut_ptr(),
-                &mut foo,
-                &mut bar,
-            );
+            unsafe {
+                ffi::PyArg_ParseTupleAndKeywords(
+                    args,
+                    kwds,
+                    c_str!("l|l").as_ptr(),
+                    #[cfg(Py_3_13)]
+                    args_names.as_ptr(),
+                    #[cfg(not(Py_3_13))]
+                    args_names.as_mut_ptr(),
+                    &mut foo,
+                    &mut bar,
+                )
+            };
 
             #[cfg(not(Py_3_13))]
-            drop(std::ffi::CString::from_raw(args_names[0]));
+            drop(unsafe { std::ffi::CString::from_raw(args_names[0]) });
             #[cfg(not(Py_3_13))]
-            drop(std::ffi::CString::from_raw(args_names[1]));
+            drop(unsafe { std::ffi::CString::from_raw(args_names[1]) });
 
-            ffi::PyLong_FromLong(foo * bar)
+            unsafe { ffi::PyLong_FromLong(foo * bar) }
         }
 
         let py_fn = PyCFunction::new_with_keywords(
